﻿//
// MainPage.xaml.cpp
// Implementation of the MainPage class.
//

#include "pch.h"    // includes wrl.h, robuffer.h, ppltasks.h
#include "MainPage.xaml.h"

using namespace TwisterCPlusPlus;

using namespace Concurrency;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::Foundation;
using namespace Windows::Graphics::Imaging;
using namespace Windows::Storage::Streams;
using namespace Windows::UI::Xaml;
using namespace Windows::UI::Xaml::Media;
using namespace Windows::UI::Xaml::Media::Imaging;

MainPage::MainPage()
{
    InitializeComponent();
    Application::Current->DebugSettings->EnableFrameRateCounter = true;
    Loaded += ref new RoutedEventHandler(this, &MainPage::OnMainPageLoaded);
}

void MainPage::OnMainPageLoaded(Object^ sender, RoutedEventArgs^ args)
{
    Uri^ uri = ref new Uri("http://www.charlespetzold.com/pw6/PetzoldJersey.jpg");
    RandomAccessStreamReference^ streamRef = RandomAccessStreamReference::CreateFromUri(uri);

    task<IRandomAccessStreamWithContentType^> (streamRef->OpenReadAsync()).
    then([](task<IRandomAccessStreamWithContentType^> thisTask)
    {
        IRandomAccessStreamWithContentType^ fileStream = thisTask.get();
        return BitmapDecoder::CreateAsync(fileStream);
    }).
    then([](task<BitmapDecoder^> thisTask)
    {
        BitmapDecoder^ decoder = thisTask.get();
        return decoder->GetFrameAsync(0);
    }).
    then([this](task<BitmapFrame^> thisTask)
    {
        BitmapFrame^ frame = thisTask.get();   

        // Save some information as fields
        width = frame->PixelWidth;
        height = frame->PixelHeight;
        xCenter = width / 2;
        yCenter = height / 2;

        return frame->GetPixelDataAsync();
    }).
    then([this](task<PixelDataProvider^> thisTask)
    {
        PixelDataProvider^ pixelProvider = thisTask.get();
        srcPixels = pixelProvider->DetachPixelData();

        // Create the WriteableBitmap 
        bitmap = ref new WriteableBitmap(width, height);

        // Set the bitmap to the Image element
        image->Source = bitmap;

        // Get access to the pixels
        IBuffer^ buffer = bitmap->PixelBuffer;

        // Obtain IBufferByteAccess
        ComPtr<IBufferByteAccess> pBufferByteAccess;
        ComPtr<IUnknown> pBuffer((IUnknown*)buffer);
        pBuffer.As(&pBufferByteAccess);
    
        // Get pointer to pixel bytes
        pBufferByteAccess->Buffer(&pDstPixels);

        // Let's go!
        CompositionTarget::Rendering += ref new EventHandler<Object^>(this, &MainPage::OnCompositionTargetRendering);
    });
}

void MainPage::OnCompositionTargetRendering(Object^ sender, Object^ args)
{
    // Get elapsed time
    TimeSpan timeSpan = dynamic_cast<RenderingEventArgs^>(args)->RenderingTime;

    // Calculate twistAngle from -180 to 180 degrees
    double t = (timeSpan.Duration % cycleDuration) / (double)cycleDuration;
    double tprime = 2 * (t < 0.5 ? t : 1 - t);
    double twistAngle = 2 * (tprime - 0.5) * 3.14159;

    for (int yDst = 0; yDst < height; yDst++)
    for (int xDst = 0; xDst < width; xDst++)
    {
        // Calculate length of point to center and angle
        int xDelta = xDst - xCenter;
        int yDelta = yDst - yCenter;
        double distanceToCenter = sqrt(xDelta * xDelta +
                                       yDelta * yDelta);
        double angleClockwise = atan2(yDelta, xDelta);

        // Calculation angle of rotation for twisting effect
        double xEllipse = xCenter * cos(angleClockwise);
        double yEllipse = yCenter * sin(angleClockwise);
        double radius = sqrt(xEllipse * xEllipse +
                             yEllipse * yEllipse);
        double fraction = max(0.0, 1 - distanceToCenter / radius);
        double twist = fraction * twistAngle;

        // Calculate the source pixel for each destination pixel
        int xSrc = (int)(xCenter + (xDst - xCenter) * cos(twist)
                                 - (yDst - yCenter) * sin(twist));
        int ySrc = (int)(yCenter + (xDst - xCenter) * sin(twist)
                                 + (yDst - yCenter) * cos(twist));
        xSrc = max(0, min(width - 1, xSrc));
        ySrc = max(0, min(height - 1, ySrc));

        // Calculate the indices
        int iDst = 4 * (yDst * width + xDst);
        int iSrc = 4 * (ySrc * width + xSrc);

        // Transfer the pixel bytes
        pDstPixels[iDst++] = srcPixels[iSrc++];
        pDstPixels[iDst++] = srcPixels[iSrc++];
        pDstPixels[iDst++] = srcPixels[iSrc++];
        pDstPixels[iDst] = srcPixels[iSrc];
    }

    // Invalidate the bitmap
    bitmap->Invalidate();
}